#!/usr/bin/python3
#-----------------------------------------------------------------------
# File    : id2par.py
# Contents: map integer identifier to a parameter setting
# Author  : Christian Borgelt
# History : 2009.08.11 file created
#           2009.08.12 turned all code into functions
#           2009.08.22 output for an out-of-range id changed
#           2009.09.01 id range shifted from [0..n-1] to [1..n]
#           2009.09.04 changed result of get_args() to dictionary
#           2009.09.05 function get_coords() split off
#           2012.06.29 grid list projection added (project())
#           2013.02.09 made compatible with Python 3 (print, range)
#-----------------------------------------------------------------------
from sys        import argv, maxsize
from os.path    import join
from subprocess import call

#-----------------------------------------------------------------------
# Parameter Selection Function
#-----------------------------------------------------------------------

def get_coords (grids, gid=-1):
    '''get_coords (grids, gid=-1)
Get the grid point coordinates that correspond to an identifier.
grids   list of parameter grid definitions
gid     identifier of the parameter grid point
returns grid point coordinates corresponding to gid'''
    if gid < 0: gid = maxsize   # find maximum index
    psum = size = 0             # initialize the grid point sum
    for g in range(len(grids)): # traverse the individual grids
        size = 1; steps = []    # initialize grid size and step list
        for p in reversed(grids[g]):
            steps.insert(0,size)# note the step size for the parameter
            if isinstance(p[0], ''.__class__): p = p[1]
            size *= len(p)      # compute the grid size from the ranges
        psum += size            # sum the grid points over the grids
        if gid < psum: break    # if the desired point is found, abort
    if gid >= psum: return psum # if not, return the number of points
    c = [g]; gid -= psum -size  # note grid index and get index in grid
    g = grids[g]                # get the selected parameter grid
    for d in range(len(g)):     # traverse the parameter definitions
        c.append(int(gid /steps[d])) # compute the grid point coordinate
        gid %= steps[d]         # and 'remove' the step position
    return c                    # return the grid point coordinates

#-----------------------------------------------------------------------

def get_args (grids, gid=-1):
    '''get_args (grids, gid=-1)
Get the parameter setting that corresponds to an identifier.
grids   list of parameter grid definitions
gid     identifier of the parameter grid point
returns parameter setting corresponding to gid'''
    if isinstance(gid, [].__class__):
        coords = gid            # use given grid point coordinates
    else:                       # get the grid point coordinates
        coords = get_coords(grids, gid)
        if isinstance(coords, (0).__class__):
            return coords       # return the total number of grid points
    args = dict()               # initialize the argument list
    for p,x in zip(grids[coords[0]],coords[1:]):
        if   isinstance(p[0],   ''.__class__):
              args[p[0]] = p[1][x]
        elif isinstance(p[x], args.__class__):
              args.update(p[x])
        else: args.update(dict(p[x]))
    return args                 # collect and return the arguments

#-----------------------------------------------------------------------

def project (grids, select):
    '''project (grids, select)
Project a parameter grid list to a parameter selection.
grids   list of parameter grid definitions
select  selection of parameter names
returns a projected parameter grid'''
    proj   = []                 # initialize the output grid list
    select = set(select)        # turn selection into a set
    for g in grids:             # traverse the individual grids
        out = []                # initialize the output grid
        for p in g:             # traverse the parameter definitions
            if isinstance(p[0], ''.__class__):
                if p[0] in select: out.append(p)
            else:               # project lists and dictionaries
                l = [[(k,d[k]) for k in d if k in select] for d in p]
                l = [dict(x) for x in l if x]
                if l: out.append(tuple(l))
        proj.append(out)        # add the projected grid
    return proj                 # return the projected grid list

#-----------------------------------------------------------------------
# Worker Execution Function
#-----------------------------------------------------------------------

def run (prog, grids, gid):
    '''run (prog, grids, gid)
Run a worker with a specific parameter setting.
prog    worker to be executed
grids   list of parameter grid definitions
gid     identifier of the parameter setting to use
returns result code of the executed program'''
    args = get_args(grids,gid)
    if   isinstance(args, (0).__class__):
        # if args is an integer, it contains the maximum id
        print(args); return args
    elif isinstance(prog, run.__class__):
        # if prog is a python function, call it directly
        return prog(**args)
    else:
        if   isinstance(prog, ''.__class__): cmd = [prog]
        elif isinstance(prog, [].__class__): cmd = prog
        else: print('error: '+str(prog)+' is not callable'); return
        for k in args.keys(): cmd.append('-' +k +str(args[k]))
        call(cmd); return 0

#-----------------------------------------------------------------------
# Internal Worker Function
#-----------------------------------------------------------------------

def worker (a,b,c):
    '''worker (a,b,c)
A dummy worker to demonstrate an internal call.'''
    print('internal worker')
    print('a: '+str(a))
    print('b: '+str(b))
    print('c: '+str(c))

#-----------------------------------------------------------------------
# Execute only if called directly
#-----------------------------------------------------------------------
if __name__ == '__main__':

    # --- worker definition
    prog = worker               # if internal worker function
    # prog = join('.','worker.py)'
                                # if external worker script/program
    # prog = [join('.','worker.py'), '-fix1', '-fix2']
                                # if there are fixed arguments
    # --- parameter grids definition
    grids = [ [ ( 'a', range(1,10)    ),
                ( 'b', range(0,5,2)   ),
                ( 'c', [1,2,3,5,7,11] ) ],
              [ ( 'a', range(2,5)     ),
                ( 'b', range(8,13,2)  ),
                ( 'c', [3,6,11,21]    ) ],
              [ ( { 'a': 0, 'b': 1 },
                  { 'a': 1, 'b': 0 },
                  { 'a': 2, 'b': 2 },
                  { 'a': 3, 'b': 3 }  ),
                ( 'c', [3,6,11,21]    ) ] ]
    # first  level: list of grid definitions
    #               (hyperboxes in the parameter space)
    # second level: list of parameter definitions
    #               (here: 3 parameters, namely a, b, and c
    # third  level: pair of parameter name and value range
    #               (in round brackets, that is, as a tuple)
    #               or dictionaries of parameter assignments

    # --- worker execution
    # argv[0] contains the script name (here: 'id2par.py')
    # argv[1] contains the first command line argument
    run(prog, grids, int(argv[1])-1 if len(argv) > 1 else -1)
